"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.uart = exports.Uart = void 0;
var stringArrayBuffer_1 = require("./helpers/stringArrayBuffer");
var jss_1 = require("jss");
var jss_preset_default_1 = require("jss-preset-default");
jss_1.default.setup((0, jss_preset_default_1.default)());
var Uart = /** @class */ (function () {
    function Uart() {
        var _this = this;
        this.styles = {
            menu: {
                color: "#b2b2b2b",
                background: "white",
                padding: "10px 16px 10px 16px",
                fontWeight: "medium",
                borderTopLeftRadius: 7,
                borderTopRightRadius: 7,
                border: "0.5px solid lightgray",
                "& .esp-tools-header-bar": {
                    display: "flex",
                    alignItems: "center",
                    justifyContent: "space-between",
                    "& *": {
                        margin: 0,
                    },
                    "& div": {
                        cursor: "pointer",
                        borderRadius: "50%",
                        display: "flex",
                        alignItems: "center",
                        justifyContent: "center",
                        "&:hover": {
                            background: "rgba(0,0,0,0.05)",
                        },
                    },
                },
            },
            items: {
                color: "#000",
                background: "white",
                padding: "4px 8px 4px 8px",
                display: "grid",
                gridTemplateColumns: "fit-content(0)",
                gridTemplateRows: "fit-content(0)",
                borderBottomRightRadius: 7,
                borderBottomLeftRadius: 7,
                border: "0.5px solid lightgray",
                borderTop: 0,
                "& p": {
                    content: "Select a connection method to pair your device",
                    gridArea: "1 / 1 / 2 / 3",
                    fontSize: 12,
                    paddingLeft: 10,
                    paddingTop: 5,
                    margin: 0,
                    color: "#7D7D7D",
                },
            },
            endpoints: {
                width: 100,
                height: 150,
                margin: "4px 0px 4px 0px",
                paddingLeft: 30,
                paddingRight: 30,
                borderRadius: 10,
                cursor: "pointer",
                display: "flex",
                flexDirection: "column",
                textAlign: "center",
                alignItems: "center",
                justifyContent: "center",
                "&:hover": {
                    background: "rgba(0,0,0,0.05)",
                },
                "& .esp-tools-icons": {
                    width: 48,
                    height: 48,
                },
                "& .esp-tools-name": {
                    fontSize: 14,
                    fontWeight: 500,
                    paddingTop: 8,
                    paddingBottom: 4,
                },
                "& .esp-tools-description": {
                    fontWeight: 300,
                    fontSize: 12,
                    color: "#666",
                },
            },
        };
        this.sentChunks = [];
        this.classList = jss_1.default.createStyleSheet(this.styles).attach();
        this.flowControl = true;
        this.debug = 1;
        this.isBusy = undefined;
        this.queue = [];
        this.endpoints = [
            {
                name: "Web Bluetooth",
                description: "Bluetooth LE devices",
                svg: '<svg xmlns="http://www.w3.org/2000/svg" width="48" height="48" viewBox="0 0 24 24"><path d="M0 0h24v24H0z" fill="none"/><path d="M17.71 7.71L12 2h-1v7.59L6.41 5 5 6.41 10.59 12 5 17.59 6.41 19 11 14.41V22h1l5.71-5.71-4.3-4.29 4.3-4.29zM13 5.83l1.88 1.88L13 9.59V5.83zm1.88 10.46L13 18.17v-3.76l1.88 1.88z" fill="#d2d2d2"/></svg>',
                isSupported: function () {
                    if (navigator.platform.indexOf("Win") >= 0 &&
                        (navigator.userAgent.indexOf("Chrome/54") >= 0 ||
                            navigator.userAgent.indexOf("Chrome/55") >= 0 ||
                            navigator.userAgent.indexOf("Chrome/56") >= 0))
                        return "Chrome <56 in Windows has navigator.bluetooth but it's not implemented properly";
                    if (window &&
                        window.location &&
                        window.location.protocol == "http:" &&
                        window.location.hostname != "localhost")
                        return "Serving off HTTP (not HTTPS) - Web Bluetooth not enabled";
                    if (navigator.bluetooth)
                        return true;
                    var iOS = /iPad|iPhone|iPod/.test(navigator.userAgent) && !window.MSStream;
                    if (iOS) {
                        return "To use Web Bluetooth on iOS you'll need the WebBLE App.\nPlease go to https://itunes.apple.com/us/app/webble/id1193531073 to download it.";
                    }
                    else {
                        return "This Web Browser doesn't support Web Bluetooth.\nPlease see https://www.espruino.com/Puck.js+Quick+Start";
                    }
                },
                connect: function (connection, callback) {
                    var _this = this;
                    var NORDIC_SERVICE = "6e400001-b5a3-f393-e0a9-e50e24dcca9e";
                    var NORDIC_TX = "6e400002-b5a3-f393-e0a9-e50e24dcca9e";
                    var NORDIC_RX = "6e400003-b5a3-f393-e0a9-e50e24dcca9e";
                    var DEFAULT_CHUNKSIZE = 20;
                    // FIND OUT CORRECT TYPES FOR THIS
                    var btServer = undefined;
                    var btService;
                    var txCharacteristic;
                    var rxCharacteristic;
                    var txDataQueue = [];
                    var flowControlXOFF = false;
                    var chunkSize = DEFAULT_CHUNKSIZE;
                    connection.close = function (callback) {
                        connection.isOpening = false;
                        if (connection.isOpen) {
                            connection.isOpen = false;
                            connection.emit("close");
                        }
                        else {
                            if (callback)
                                callback(null);
                        }
                        if (btServer) {
                            btServer.disconnect();
                            btServer = undefined;
                            txCharacteristic = undefined;
                            rxCharacteristic = undefined;
                        }
                    };
                    connection.write = function (data, callback) {
                        var writeChunk = function () {
                            if (flowControlXOFF) {
                                // flow control - try again later
                                setTimeout(writeChunk, 50);
                                return;
                            }
                            var chunk;
                            var txItem = txDataQueue[0];
                            if (txItem.data.length <= chunkSize) {
                                chunk = txItem.data;
                                txItem.data = undefined;
                            }
                            else {
                                chunk = txItem.data.substr(0, chunkSize);
                                txItem.data = txItem.data.substr(chunkSize);
                            }
                            connection.txInProgress = true;
                            console.log("<UART>", "Sending " + JSON.stringify(chunk));
                            txCharacteristic
                                .writeValue((0, stringArrayBuffer_1.str2ab)(chunk))
                                .then(function () {
                                console.log("<UART>", "Sent");
                                if (!txItem.data) {
                                    txDataQueue.shift(); // remove this element
                                    if (txItem.callback)
                                        txItem.callback();
                                }
                                connection.txInProgress = false;
                                writeChunk();
                            })
                                .catch(function (error) {
                                console.log("<UART>", "SEND ERROR: " + error);
                                txDataQueue = [];
                                connection.close();
                            });
                        };
                        if (data)
                            txDataQueue.push({
                                data: data,
                                callback: callback,
                                maxLength: data.length,
                            });
                        if (connection.isOpen && !connection.txInProgress)
                            writeChunk();
                    };
                    navigator.bluetooth
                        .requestDevice({
                        filters: [
                            { namePrefix: "Puck.js" },
                            { namePrefix: "Pixl.js" },
                            { namePrefix: "MDBT42Q" },
                            { namePrefix: "Bangle" },
                            { namePrefix: "RuuviTag" },
                            { namePrefix: "iTracker" },
                            { namePrefix: "Thingy" },
                            { namePrefix: "Espruino" },
                            { services: [NORDIC_SERVICE] },
                        ],
                        optionalServices: [NORDIC_SERVICE],
                    })
                        .then(function (device) {
                        console.log("<UART>", "Device Name:       " + device.name);
                        console.log("<UART>", "Device ID:         " + device.id);
                        // Was deprecated: Should use getPrimaryServices for this in future
                        //log('BT>  Device UUIDs:      ' + device.uuids.join('\n' + ' '.repeat(21)));
                        device.addEventListener("gattserverdisconnected", function () {
                            console.log("<UART>", "Disconnected (gattserverdisconnected)");
                            connection.close();
                        });
                        return device.gatt.connect();
                    })
                        .then(function (server) {
                        console.log("<UART>", "Connected");
                        btServer = server;
                        return server.getPrimaryService(NORDIC_SERVICE);
                    })
                        .then(function (service) {
                        console.log("<UART>", "Got service");
                        btService = service;
                        return btService.getCharacteristic(NORDIC_RX);
                    })
                        .then(function (characteristic) {
                        rxCharacteristic = characteristic;
                        console.log("<UART>", "RX characteristic:" + JSON.stringify(rxCharacteristic));
                        rxCharacteristic.addEventListener("characteristicvaluechanged", function (event) {
                            var dataview = event.target.value;
                            if (dataview.byteLength > chunkSize) {
                                console.log("<UART>", "Received packet of length " +
                                    dataview.byteLength +
                                    ", increasing chunk size");
                                chunkSize = dataview.byteLength;
                            }
                            if (_this.flowControl) {
                                for (var i = 0; i < dataview.byteLength; i++) {
                                    var ch = dataview.getUint8(i);
                                    if (ch == 17) {
                                        // XON
                                        console.log("<UART>", "XON received => resume upload");
                                        flowControlXOFF = false;
                                    }
                                    if (ch == 19) {
                                        // XOFF
                                        console.log("<UART>", "XOFF received => pause upload");
                                        flowControlXOFF = true;
                                    }
                                }
                            }
                            var str = (0, stringArrayBuffer_1.ab2str)(dataview.buffer);
                            console.log("<UART>", "Received " + JSON.stringify(str));
                            connection.emit("data", str);
                        });
                        return rxCharacteristic.startNotifications();
                    })
                        .then(function () {
                        return btService.getCharacteristic(NORDIC_TX);
                    })
                        .then(function (characteristic) {
                        txCharacteristic = characteristic;
                        console.log("<UART>", "TX characteristic:" + JSON.stringify(txCharacteristic));
                    })
                        .then(function () {
                        connection.txInProgress = false;
                        connection.isOpen = true;
                        connection.isOpening = false;
                        _this.isBusy = false;
                        _this.queue = [];
                        callback(connection);
                        connection.emit("open");
                        // if we had any writes queued, do them now
                        connection.write();
                    })
                        .catch(function (error) {
                        console.log("<UART>", "ERROR: " + error);
                        connection.close();
                    });
                    return connection;
                },
            },
            {
                name: "Web Serial",
                description: "USB connected devices",
                svg: '<svg xmlns="http://www.w3.org/2000/svg" width="48" height="48" viewBox="0 0 24 24"><path d="M0 0h24v24H0z" fill="none"/><path d="M15 7v4h1v2h-3V5h2l-3-4-3 4h2v8H8v-2.07c.7-.37 1.2-1.08 1.2-1.93 0-1.21-.99-2.2-2.2-2.2-1.21 0-2.2.99-2.2 2.2 0 .85.5 1.56 1.2 1.93V13c0 1.11.89 2 2 2h3v3.05c-.71.37-1.2 1.1-1.2 1.95 0 1.22.99 2.2 2.2 2.2 1.21 0 2.2-.98 2.2-2.2 0-.85-.49-1.58-1.2-1.95V15h3c1.11 0 2-.89 2-2v-2h1V7h-4z" fill="#d2d2d2"/></svg>',
                isSupported: function () {
                    if (!navigator.serial)
                        return "No navigator.serial - Web Serial not enabled";
                    if (window &&
                        window.location &&
                        window.location.protocol == "http:" &&
                        window.location.hostname != "localhost")
                        return "Serving off HTTP (not HTTPS) - Web Serial not enabled";
                    return true;
                },
                connect: function (connection, callback) {
                    var _this = this;
                    var serialPort;
                    var disconnected = function () {
                        connection.isOpening = false;
                        if (connection.isOpen) {
                            console.log("<UART>", "Disconnected");
                            connection.isOpen = false;
                            connection.emit("close");
                        }
                    };
                    // TODO: Pass USB vendor and product ID filter when supported by Chrome.
                    navigator.serial
                        .requestPort()
                        .then(function (port) {
                        console.log("<UART>", "Connecting to serial port");
                        serialPort = port;
                        return port.open({ baudRate: 115200 });
                    })
                        .then(function () {
                        var readLoop = function () {
                            var reader = serialPort.readable.getReader();
                            // FIND OUT CORRECT TYPES FOR THIS
                            reader.read().then(function (_a) {
                                var value = _a.value, done = _a.done;
                                reader.releaseLock();
                                if (value) {
                                    var str = (0, stringArrayBuffer_1.ab2str)(value.buffer);
                                    console.log("<UART>", "Received " + JSON.stringify(str));
                                    connection.emit("data", str);
                                }
                                if (done) {
                                    disconnected();
                                }
                                else {
                                    readLoop();
                                }
                            });
                        };
                        readLoop();
                        console.log("<UART>", "Serial connected. Receiving data...");
                        connection.txInProgress = false;
                        connection.isOpen = true;
                        connection.isOpening = false;
                        callback(connection);
                    })
                        .catch(function (error) {
                        console.log("<UART>", "ERROR: " + error);
                        disconnected();
                    });
                    connection.close = function (callback) {
                        if (serialPort) {
                            serialPort.close();
                            serialPort = undefined;
                        }
                        disconnected();
                    };
                    connection.write = function (data, callback) {
                        var writer = serialPort.writable.getWriter();
                        // TODO: progress?
                        writer
                            .write((0, stringArrayBuffer_1.str2ab)(data))
                            .then(function () {
                            callback === null || callback === void 0 ? void 0 : callback();
                        })
                            .catch(function (error) {
                            console.log("<UART>", "SEND ERROR: " + error);
                            _this.closeSerial();
                        });
                        writer.releaseLock();
                    };
                    return connection;
                },
            },
        ];
        this.getWrittenData = function () {
            var str_chunks = _this.sentChunks.join("");
            return new Promise(function (resolve) { return resolve(str_chunks); });
        };
        this.connect = function (callback) {
            var connection = {
                on: function (evt, cb) {
                    this["on" + evt] = cb;
                },
                emit: function (evt, data) {
                    if (this["on" + evt])
                        this["on" + evt](data);
                },
                isOpen: false,
                isOpening: true,
                txInProgress: false,
            };
            _this.connection.hadData = false;
            // modal
            var e = document.createElement("div");
            e.setAttribute("style", "position:absolute;top:0px;left:0px;right:0px;bottom:0px;opacity:0.5;z-index:100;background:black;");
            // menu
            var menu = document.createElement("div");
            menu.setAttribute("style", "position:absolute;left:50%;top:50%;transform:translate(-50%,-50%);font-family: Sans-Serif;z-index:101;");
            var menutitle = document.createElement("div");
            menutitle.classList.add(_this.classList.classes.menu);
            var menuContent = document.createElement("div");
            menuContent.classList.add("esp-tools-header-bar");
            var menuTitle = document.createElement("p");
            menuTitle.innerText = "Connect";
            menuContent.appendChild(menuTitle);
            var menuClose = document.createElement("div");
            menuClose.innerHTML =
                '<svg id="esp-tools-close-modal" stroke="currentColor" fill="currentColor" stroke-width="0" viewBox="0 0 24 24" height="1em" width="1em" xmlns="http://www.w3.org/2000/svg"><path fill="none" stroke="#000" stroke-width="2" d="M7,7 L17,17 M7,17 L17,7"></path></svg>';
            menuContent.appendChild(menuClose);
            menutitle.appendChild(menuContent);
            menu.appendChild(menutitle);
            var items = document.createElement("div");
            items.classList.add(_this.classList.classes.items);
            var p = document.createElement("p");
            p.innerText = "Select a connection method to pair your device";
            items.appendChild(p);
            menu.appendChild(items);
            // FIND OUT CORRECT TYPES FOR THIS
            _this.endpoints.forEach(function (endpoint) {
                var supported = endpoint.isSupported();
                if (supported !== true)
                    console.log("<UART>", endpoint.name + " not supported, " + supported);
                var ep = document.createElement("div");
                ep.classList.add(_this.classList.classes.endpoints);
                ep.innerHTML =
                    '<div class="esp-tools-icons">' +
                        endpoint.svg +
                        "</div>" +
                        '<div class="esp-tools-name">' +
                        endpoint.name +
                        "</div>" +
                        '<div class="esp-tools-description">' +
                        endpoint.description +
                        "</div>";
                ep.onclick = function (evt) {
                    connection = endpoint.connect(connection, callback);
                    evt.preventDefault();
                    document.body.removeChild(menu);
                    document.body.removeChild(e);
                };
                items.appendChild(ep);
            });
            menuClose.onclick = function () {
                document.body.removeChild(menu);
                document.body.removeChild(e);
                connection.isOpening = false;
                if (connection.isOpen) {
                    connection.isOpen = false;
                }
                else {
                    if (callback)
                        callback(null);
                }
            };
            document.body.appendChild(e);
            document.body.appendChild(menu);
            return connection;
        };
    }
    Uart.prototype.log = function (level, s) {
        if (level <= this.debug)
            console.log("<UART> " + s);
    };
    Uart.prototype.checkIfSupported = function () {
        var anySupported = false;
        // FIND OUT CORRECT TYPES FOR THIS
        this.endpoints.forEach(function (endpoint) {
            var supported = endpoint.isSupported();
            if (supported === true)
                anySupported = true;
            else
                console.log("<UART>", endpoint.name + " not supported, " + supported);
        });
        return anySupported;
    };
    Uart.prototype.eval = function (expr, cb) {
        if (!this.checkIfSupported())
            return false;
        if (this.isBusy) {
            console.log("<UART>", "Busy - adding eval to queue");
            this.queue.push({ type: "eval", expr: expr, cb: cb });
            return false;
        }
        this.write("\x10eval(process.env.CONSOLE).println(JSON.stringify(" + expr + "))\n", function (d) {
            try {
                var json = JSON.parse(d.trim());
                cb(json, "success");
            }
            catch (e) {
                console.log("<UART>", "Unable to decode " + JSON.stringify(d) + ", got " + e.toString());
                cb(null, "failed");
            }
        }, true /*callbackNewline*/);
        return true;
    };
    Uart.prototype.write = function (data, callback, callbackNewline) {
        var _this = this;
        if (!this.checkIfSupported())
            return;
        if (this.isBusy) {
            console.log("<UART>", "Busy - adding write to queue");
            this.queue.push({
                type: "write",
                data: data,
                callback: callback,
                callbackNewline: callbackNewline,
            });
            return;
        }
        // FIND OUT CORRECT TYPES FOR THIS
        var cbTimeout;
        var onWritten = function () {
            if (callbackNewline) {
                _this.connection.cb = function () {
                    var newLineIdx = _this.connection.received.indexOf("\n");
                    if (newLineIdx >= 0) {
                        var l = _this.connection.received.substr(0, newLineIdx);
                        _this.connection.received = _this.connection.received.substr(newLineIdx + 1);
                        _this.connection.cb = undefined;
                        if (cbTimeout)
                            clearTimeout(cbTimeout);
                        cbTimeout = undefined;
                        if (callback)
                            callback(l);
                        _this.isBusy = false;
                        _this.handleQueue();
                    }
                };
            }
            // wait for any received data if we have a callback...
            var maxTime = 300; // 30 sec - Max time we wait in total, even if getting data
            var dataWaitTime = callbackNewline
                ? 100 /*10 sec  if waiting for newline*/
                : 3; /*300ms*/
            var maxDataTime = dataWaitTime; // max time we wait after having received data
            var timeout = function () {
                cbTimeout = undefined;
                if (maxTime)
                    maxTime--;
                if (maxDataTime)
                    maxDataTime--;
                if (_this.connection.hadData)
                    maxDataTime = dataWaitTime;
                if (maxDataTime && maxTime) {
                    cbTimeout = setTimeout(timeout, 100);
                }
                else {
                    _this.connection.cb = undefined;
                    if (callbackNewline)
                        console.log("<UART>", "write waiting for newline timed out");
                    if (callback)
                        callback(_this.connection.received);
                    _this.isBusy = false;
                    _this.handleQueue();
                    _this.connection.received = "";
                }
                _this.connection.hadData = false;
            };
            cbTimeout = setTimeout(timeout, 100);
        };
        if (this.connection &&
            (this.connection.isOpen || this.connection.isOpening)) {
            if (!this.connection.txInProgress)
                this.connection.received = "";
            this.isBusy = true;
            return this.connection.write(data, onWritten);
        }
        // FIND OUT CORRECT TYPES FOR THIS
        this.connection = this.connect(function (uart) {
            if (!uart) {
                _this.connection = undefined;
                if (callback)
                    callback(null);
                return;
            }
            _this.connection.received = "";
            _this.connection.on("data", function (d) {
                _this.connection.received += d;
                _this.connection.hadData = true;
                if (_this.connection.cb)
                    _this.connection.cb(d);
            });
            _this.connection.on("close", function (d) {
                _this.connection = undefined;
            });
            _this.isBusy = true;
            _this.connection.write(data, onWritten);
        });
    };
    Uart.prototype.modal = function (callback) {
        var e = document.createElement("div");
        e.setAttribute("style", "position:absolute;top:0px;left:0px;right:0px;bottom:0px;opacity:0.5;z-index:100;background:black;");
        e.innerHTML =
            '<div style="position:absolute;left:50%;top:50%;transform:translate(-50%,-50%);font-family: Sans-Serif;font-size:400%;color:white;">Click to Continue...</div>';
        e.onclick = function (evt) {
            callback();
            evt.preventDefault();
            document.body.removeChild(e);
        };
        document.body.appendChild(e);
    };
    Uart.prototype.handleQueue = function () {
        if (!this.queue.length)
            return;
        var q = this.queue.shift();
        console.log("<UART>", "Executing " + JSON.stringify(q) + " from queue");
        if (q.type == "eval")
            this.eval(q.expr, q.cb);
        else if (q.type == "write")
            this.write(q.data, q.callback, q.callbackNewline);
        else
            console.log("<UART>", "Unknown queue item " + JSON.stringify(q));
    };
    return Uart;
}());
exports.Uart = Uart;
var uart = new Uart();
exports.uart = uart;
